; parser for converting GCC-XML file into IR

(in-package "FRONTEND")

; the id of the root namespace
(defparameter +root-namespace-id+ "_1")

; the virtual id of the ellipses type
(defparameter +ellipsis-type-id+ "_ELLIPSIS")

; the id of the constant integer type
(defparameter +const-integer-type-id+ nil)

; map of handlers for driving the parser
(defparameter +handler-map+
  '((:|ArrayType|        . handle-array-type)
    (:|Class|            . handle-class)
    (:|Constructor|      . handle-constructor)
    (:|Converter|        . handle-converter)
    (:|CvQualifiedType|  . handle-cv-qualified-type)
    (:|Destructor|       . handle-destructor)
    (:|Enumeration|      . handle-enumeration)
    (:|Function|         . handle-function)
    (:|FunctionType|     . handle-function-type)
    (:|FundamentalType|  . handle-fundamental-type)
    (:|Method|           . handle-function)
    (:|MethodType|       . handle-function-type)
    (:|Namespace|        . handle-namespace)
    (:|NamespaceAlias|   . handle-namespace-alias)
    (:|OperatorFunction| . handle-operator)
    (:|OperatorMethod|   . handle-operator)
    (:|PointerType|      . handle-pointer-type)
    (:|ReferenceType|    . handle-pointer-type)
    (:|Struct|           . handle-class)
    (:|Typedef|          . handle-typedef)
    (:|Union|            . handle-union)))

; types that are defined by a lexical context
(defparameter +defined-types+
  '(:|ArrayType| :|Class| :|CvQualifiedType| :|Enumeration| 
    :|FunctionType| :|FundamentalType| :|MethodType| :|Namespace| 
    :|NamespaceAlias| :|OffsetType| :|PointerType| :|ReferenceType| 
    :|Struct| :|Typedef| :|Union|))

; types that are allocated by a lexical context
(defparameter +allocated-types+
  '(:|Constructor| :|Converter| :|Destructor| :|Field| :|Function|
    :|Method| :|Variable|))

; types that require an extra indirection (through :|types|)
; when building the IR from the XML tree
(defparameter +indirect-allocated-types+
  '(:|Field| :|Variable|))

; lookup the handler for a given XML element
(defun lookup-handler (element) 
  (cdr (assoc element +handler-map+)))

; parse XML file and create IR
(defun parse-gccxml-output (path)
  (let ((dom (read-xml-nodes path)))
    (convert-xml-nodes dom)))

; read XML file and return lxml DOM 
(defun read-xml-nodes (path)
  (s-xml:parse-xml-file path))

; convert from XML tree to IR graph
(defun convert-xml-nodes (dom)
  (let ((ir-index (make-hash-table :test #'equal))
	(xml-index (index-xml-nodes dom))
	(edge-queue (make-array 1 :fill-pointer 0)))
    (handle-special-nodes xml-index ir-index)
    (dolist (node dom) (handle-xml-node node xml-index ir-index edge-queue))
    (build-graph ir-index edge-queue)))

; helper function for finding the "const int" type
(defun is-const-int-type (node xml-index)
  (when (and (eq (get-element-name node) ':|CvQualifiedType|)
	     (equal (get-element-attribute node ':|const|) "1"))
    (let ((tgt (gethash (get-element-attribute node ':|type|) xml-index)))
      (when (equal (get-element-attribute tgt ':|name|) "int")
	t))))

; insert some special nodes into the indices
(defun handle-special-nodes (xml-index ir-index)
  (setf (gethash +ellipsis-type-id+ ir-index) (make-instance 'ellipsis-type))
  (maphash #'(lambda (k v)
	       (when (is-const-int-type v xml-index)
		 (setf +const-integer-type-id+ k)))
	   xml-index))

; index xml nodes by their id attribute
(defun index-xml-nodes (dom)
  (let ((index (make-hash-table :test #'equal)))
    (dolist (node dom index)
      (let ((id (get-element-id node)))
	(when id
	  (setf (gethash id index) node))))))

; create IR node for XML node, enter it into index
(defun handle-xml-node (node xml-index ir-index queue)
  (let* ((elname (get-element-name node))
	 (handler (lookup-handler elname)))
    (when handler
      (funcall handler node xml-index ir-index queue))))

; given an index of IR nodes and a queue of edges, build a graph
(defun build-graph (ir-index edge-queue)
  (map nil #'(lambda (queued-edge)
	      (destructuring-bind (from to kind slots) queued-edge
		(add-edge (gethash from ir-index) (gethash to ir-index) 
			  kind slots))) edge-queue)
  (let ((root (make-instance 'library))
	(rns (gethash +root-namespace-id+ ir-index)))
    (add-edge root rns 'defines nil)
    (add-edge root rns 'allocates (list (cons 'name "::")))
    root))

; macros to make it easier to define handlers
(defmacro with-current-state (ir-node xml-node edge-queue &rest body) 
  `(flet ((fill-node-slot (slot attr def)
	    (fill-slot-from-xml ,ir-node slot ,xml-node attr def))
	 (queue-node-edges (attr etype)
	    (queue-edges-from-xml ,xml-node attr etype ,edge-queue))
	 (fill-node-annotation (name attr def)
	    (fill-annotation-from-xml ,ir-node name ,xml-node attr def))
	 (get-current-id-list (attr)
	     (get-id-list ,xml-node attr)))
    ,@body))

; handle ArrayType XML elements
(defun handle-array-type (node xml-index ir-index edge-queue)
  (let ((irn (make-and-index-node 'array-type node ir-index)))
    (with-current-state irn node edge-queue
      (fill-node-slot 'size ':|max| 0)
      (queue-node-edges ':|type| 'extends))))

; handle Class XML elements
(defun handle-class (node xml-index ir-index edge-queue)
  (let ((irn (make-and-index-node 'class-type node ir-index)))
    (with-current-state irn node edge-queue
      (fill-node-slot 'name ':|name| nil)
      (handle-class-bases irn node xml-index edge-queue)
      (handle-members irn node xml-index ir-index edge-queue))))

; handle the members of lexical contexts
(defun handle-members (irn node xml-index ir-index edge-queue)
  (let ((id (get-element-id node)))
    (with-current-state irn node edge-queue
      (dolist (mem (get-id-list node ':|members|))
	(let* ((mnode (gethash mem xml-index))
	       (mtype (get-element-name mnode))
	       (memid (get-element-id mnode)))
	  (cond ((member mtype +defined-types+) 
		 (queue-edges id memid 'defines edge-queue))
		((member mtype +allocated-types+)
		 (let ((name (get-element-attribute mnode ':|name|))
		       (init (get-element-attribute mnode ':|init|))
		       (slots nil))
		   (when name (push (cons 'name name) slots))
		   (when init (push (cons 'value init) slots))
		   (if (member mtype +indirect-allocated-types+)
		     (queue-edges id (get-element-attribute mnode ':|type|)
				  'allocates edge-queue slots)
		     (queue-edges id memid 'allocates edge-queue slots))))))))))

; queue edges for the members of a class
(defun handle-class-bases (irn node xml-index edge-queue)
  (dolist (base (get-element-children node))
    (queue-edges (get-element-id node) (get-element-attribute base ':|type|)
		 'extends edge-queue)))

; handle general functions and methods
(defun handle-function-base (node xml-index ir-index edge-queue role)
  (let ((irn (make-and-index-node 'function-type node ir-index)))
    (with-current-state irn node edge-queue
      (set-annotation irn "function-role" role)
      (handle-function-arguments irn node xml-index edge-queue))))

; lookup the type of an argument, handling Ellipsis elements
(defun lookup-argument-type-id (node)
  (let ((name (get-element-name node)))
    (if (eq ':|Ellipsis| name)
      +ellipsis-type-id+
      (get-element-attribute node ':|type|))))

; handle a function's arguments and return type
(defun handle-function-arguments (irn node xml-index edge-queue)
  (queue-edges (get-element-id node) (get-element-attribute node ':|returns|)
	       'returns edge-queue)
  (dolist (arg (get-element-children node))
    (queue-edges (get-element-id node) (lookup-argument-type-id arg)
		 'receives edge-queue)))

; handle Constructor XML elements
(defun handle-constructor (node xml-index ir-index edge-queue)
  (handle-function-base node xml-index ir-index edge-queue "constructor"))

; handle Converter XML elements
(defun handle-converter (node xml-index ir-index edge-queue)
  (handle-function-base node xml-index ir-index edge-queue "converter"))

; handle CvQualifiedType XML elements
(defun handle-cv-qualified-type (node xml-index ir-index edge-queue)
  (let ((irn (make-and-index-node 'qualified-type node ir-index)))
    (with-current-state irn node edge-queue
      (fill-node-slot 'const ':|const| 0)
      (queue-node-edges ':|type| 'extends))))

; handle Destructor XML elements
(defun handle-destructor (node xml-index ir-index edge-queue)
  (handle-function-base node xml-index ir-index edge-queue "destructor"))

; handle Enumeration XML elements
(defun handle-enumeration (node xml-index ir-index edge-queue)
  (let ((irn (make-and-index-node 'enum-type node ir-index)))
    (with-current-state irn node edge-queue
      (fill-node-slot 'name ':|name| nil)
      (handle-enumeration-values irn node xml-index edge-queue))))

; handle EnumValue XML elements
(defun handle-enumeration-values (irn node xml-index edge-queue)
  (dolist (val (get-element-children node))
    (queue-edges (get-element-id node) +const-integer-type-id+
		 'allocates edge-queue)))

; handle Function XML elements
(defun handle-function (node xml-index ir-index edge-queue)
  (handle-function-base node xml-index ir-index edge-queue "plain"))

; handle FunctionType XML elements
(defun handle-function-type (node xml-index ir-index edge-queue)
  (handle-function-base node xml-index ir-index edge-queue "plain"))

; handle FundamentalType XML elements
(defun handle-fundamental-type (node xml-index ir-index edge-queue)
  (let ((irn (make-and-index-node 'fundamental-type node ir-index)))
    (with-current-state irn node edge-queue
      (fill-node-slot 'name ':|name| nil))))

; handle Namespace XML elements
(defun handle-namespace (node xml-index ir-index edge-queue)
  (let ((irn (make-and-index-node 'namespace-type node ir-index)))
    (with-current-state irn node edge-queue
      (fill-node-slot 'name ':|name| nil)
      (handle-members irn node xml-index ir-index edge-queue))))

; handle NamespaceAlias XML elements
(defun handle-namespace-alias (node xml-index ir-index edge-queue)
  (let ((irn (make-and-index-node 'alias-type node ir-index)))
    (with-current-state irn node edge-queue
        (fill-node-slot 'name ':|name| nil)
	(queue-node-edges ':|namespace| 'extends))))

; handle OperatorFunction, OperatorMethod XML elements
(defun handle-operator (node xml-index ir-index edge-queue)
  (handle-function-base node xml-index ir-index edge-queue "operator"))

; handle PointerType XML elements
(defun handle-pointer-type (node xml-index ir-index edge-queue)
  (let ((irn (make-and-index-node 'pointer-type node ir-index)))
    (with-current-state irn node edge-queue
      (queue-node-edges ':|type| 'extends))))

; handle Typedef XML elements
(defun handle-typedef (node xml-index ir-index edge-queue)
  (let ((irn (make-and-index-node 'alias-type node ir-index)))
    (with-current-state irn node edge-queue
      (fill-node-slot 'name ':|name| nil)
      (queue-node-edges ':|type| 'extends))))

; handle Union XML elements
(defun handle-union (node xml-index ir-index edge-queue)
  (let ((irn (make-and-index-node 'union-type node ir-index)))
    (with-current-state irn node edge-queue
      (fill-node-slot 'name ':|name| nil)
      (handle-members irn node xml-index ir-index edge-queue))))
