; utility routines for the front-end

(in-package "FRONTEND")

; get the element after an element in the sequence
(defun element-after (seq key)
  (let ((pos (position key seq)))
    (when pos
      (elt seq (+ 1 pos)))))

; get the ID attribute of an XML node
(defun get-element-id (node)
  (get-element-attribute node ':|id|))

; return the name of the xml element in lxml format
(defun get-element-name (node)
  (if (not (listp node))
    node
    (if (not (listp (car node)))
      (car node)
      (caar node))))

; return value of attribute of node in lxml format
(defun get-element-attribute (node attr)
  (if (listp (car node))
    (element-after (car node) attr)
    (element-after node attr)))

; return the given attribute value as in integer
(defun get-element-attribute-int (node attr)
  (parse-integer (get-element-attribute node attr)))

; return the children of a node in lxml format
(defun get-element-children (node)
  (when (listp (car node))
    (remove-if #'stringp (cdr node))))

; copy the contents of node attributes into node slots
(defun fill-slot-from-xml (irn slot node attr def)
  (let ((value (get-element-attribute node attr)))
    (if value
      (setf (slot-value irn slot) value)
      (setf (slot-value irn slot) def))))

; copy the contents of node attributes into node annotations
(defun fill-annotation-from-xml (irn name node attr def)
  (let ((value (get-element-attribute node attr)))
    (if value
      (set-annotation irn name value)
      (set-annotation irn name def))))

; get a list of ids specified by an attribute in a node
(defun get-id-list (node attr)
  (let ((value (get-element-attribute node attr)))
    (when value
      (split-sequence #\Space value))))

; add an edge to the edge queue using target's id
(defun queue-edges (from to kind queue &optional slots)
  (if (listp to)
    (dolist (id to) (vector-push-extend (list from id kind slots) queue))
    (vector-push-extend (list from to kind slots) queue)))

; utility function to get ids and queue edges at once
(defun queue-edges-from-xml (node attr kind queue &optional slots)
  (let ((id (get-element-id node))
	(targets (get-id-list node attr)))
    (queue-edges id targets kind queue slots)))

; create a node with given type and enter into index
(defun make-and-index-node (ir-type node index)
  (let ((irn (make-instance ir-type)))
    (setf (gethash (get-element-id node) index) irn)
    irn))

; print the IR in topological order from the root
(defun print-ir (root)
  (when (and root (not (node-mark root)))
    (print-node root)
    (mark-node root)
    (dolist (edge (node-edges root))
      (print-edge edge))
    (dolist (edge (node-edges root))
      (print-ir (edge-target edge)))))

(defmethod print-node ((root node))
  (format t "Generic Node ~A~%" root))

(defmethod print-node ((root cpp-type))
  (format t "Cpp-Type ~A~%" (cpp-type-name root)))

(defmethod print-edge ((edge edge))
  (format t "    ~A ~A~%" edge (edge-target edge)))

(defmethod print-edge ((edge allocates))
  (format t "    ~A ~A named ~A~%" edge (edge-target edge)
	  (allocates-edge-name edge)))
